-- ShortestPath.sql
-- Creates a demo database named dbShortestPath, creates a
-- demo graph in table tblGraph, creates a stored procedure
-- named usp_ShortestPath that performs shortest path analysis
-- using a variation of Dijkstra's Shortest Path algorithm.

use master
go

if exists(select * from sys.sysdatabases where name='dbShortPathDemo')
drop database dbShortPathDemo
go

create database dbShortPathDemo
go

use dbShortPathDemo
go

create table tblGraph
(
fromNode bigint not null,
toNode bigint not null,
edgeWeight real not null -- float(24)
)
go

set nocount on
insert into tblGraph values(111,222,1.0)
insert into tblGraph values(111,555,1.0)
insert into tblGraph values(222,111,2.0)
insert into tblGraph values(222,333,1.0)
insert into tblGraph values(222,555,1.0)
insert into tblGraph values(333,666,1.0)
insert into tblGraph values(333,888,1.0)
insert into tblGraph values(444,888,1.0)
insert into tblGraph values(555,111,2.0)
insert into tblGraph values(555,666,1.0)
insert into tblGraph values(666,333,2.0)
insert into tblGraph values(666,777,1.0)
insert into tblGraph values(777,444,2.0)
insert into tblGraph values(777,888,1.0)
go

create nonclustered index idxTblGraphFromNode on tblGraph(fromNode)
go
create nonclustered index idxTblGraphToNode on tblGraph(toNode)
go

-- ------------------------------------------------

create table tblNodes
(
nodeID bigint primary key
)
go

insert into tblNodes
select distinct fromNode from tblGraph
union
select distinct toNode from tblGraph
go

-- -----------------------------------------------

create table tblAlgorithmData
(
nodeID bigint not null,
dist real not null,      -- distance from startNode to node with nodeID
pred bigint not null,
inQueue bit not null
)
go

create index idxTblAlgorithmDataNodeID on tblAlgorithmData(nodeID)
create index idxTblAlgorithmDataDist on tblAlgorithmData(dist)
create index idxTblAlgorithmDataPred on tblAlgorithmData(pred)
create index idxTblAlgorithmDataInQueue on tblAlgorithmData(inQueue)
go

-- -----------------------------------------------
--drop procedure usp_ShortestPath
--go
-- -----------------------------------------------

create procedure usp_ShortestPath
 @startNode bigint,
 @endNode bigint,
 @path varchar(5000) output,
 @totDist real output
as
begin
 -- verify startNode and endNode exist in tblNodes
 --if not exists(select nodeID from tblNodes where @startNode = nodeID)
 --begin
 -- RAISERROR('START NODE DOES NOT EXIST', 9, 1)
 -- return
 --end
 
 --if not exists(select nodeID from tblNodes where @endNode = nodeID)
 --begin
 -- RAISERROR('END NODE DOES NOT EXIST', 9, 2)
 -- return
 --end
 
 -- initialize tblAlgorithmData  with startNode info
 truncate table tblAlgorithmData -- clear data from last call
 insert into tblAlgorithmData
 values (@startNode, 0.0, -1, 1) -- nodeID, dist, pred, inQueue

 declare @u bigint         -- node with smallest dist and inQueue = 1
 declare @d real = 0.0     -- distance between @u (as a fromNode) and each neighbor v (as a toNode)
 declare @done bit = 0     -- dummy loop control variable
 
 while @done = 0  -- could add a counter to limit huge runs, and other stop heuristics
 begin

  -- set @u to node with smallest distance still in the virtual queue
   select top 1 @u = nodeID, @d = dist from tblAlgorithmData
   where inQueue = 1
   order by dist asc

   if @d = 2147483647.0 break  -- no further progress possible
   if @u = @endNode break      -- we hit the endNode
   
   -- insert into tblAlgorithmData all neighbors of @u if this is the first time seen
   insert into tblAlgorithmData
   select toNode, 2147483647.0, -1, 1 from tblGraph where fromNode = @u
   and not exists (select * from tblAlgorithmData where nodeID = toNode)
      
   update tblAlgorithmData set inQueue = 0 where nodeID = @u -- 'remove u from Q'

   update tblAlgorithmData
   set dist = @d + tblGraph.edgeWeight, pred = @u
   from tblAlgorithmData inner join tblGraph on tblAlgorithmData.nodeID = tblGraph.toNode
   where tblGraph.fromNode = @u and tblAlgorithmData.inQueue = 1
    and (@d + tblGraph.edgeWeight) < tblAlgorithmData.dist

 end -- loop
  
 if (@u != @endNode)
 begin
  set @path = 'NOT REACHABLE'
  set @totDist = -1.0
 end
 else
 begin
  set @path = ''
  declare @currNode bigint
  set @currNode = @endNode
 
  while @currNode != @startNode
  begin
   set @path = @path + cast(@currNode as varchar(19)) + ';'
   set @currNode = (select pred from tblAlgorithmData where nodeID = @currNode)
  end
  
  set @path = @path + cast(@startNode as varchar(19))
  select @totDist = dist from tblAlgorithmData where nodeID = @endNode
 end -- else
 
end -- usp_ShortestPath definition
go

-- -------------------------------------------------------------------------------------

declare @startNode bigint
declare @endNode bigint
declare @path varchar(5000)
declare @totDist real

set @startNode = 222
set @endNode = 444

exec usp_ShortestPath @startNode, @endNode, @path output, @totDist output

select @path
select @totDist

select * from tblAlgorithmData

-- ================================================================================================
-- end ShortestPath.sql